/*  This program is free software; you can redistribute it and/or modify
*  it under the terms of the GNU General Public License as published by
*  the Free Software Foundation; either version 2 of the License, or
*  (at your option) any later version.
*
*  This program is distributed in the hope that it will be useful,
*  but WITHOUT ANY WARRANTY; without even the implied warranty of
*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*  GNU General Public License for more details.
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
// (C) PicoPeta Simputers Pvt. Ltd. (2003)
// Author: Santhosh N

#include "clipboard.h"
#include "display.h"
#include "eventloop.h"
#include <stdlib.h>
#include <string.h>
#include <X11/Xatom.h>

px_clipboard* clipboard_create()
{
	px_clipboard* clip = (px_clipboard*)malloc(sizeof(px_clipboard));
	Display* display = display_get_Xdisplay();
	
	clip->text = NULL;
	clip->selection_owner = NULL;
	
	clip->win = XCreateWindow(display, RootWindow(display, DefaultScreen(display)), 0, 0, 1, 1, 0,
			0, InputOnly, DefaultVisual(display, DefaultScreen(display)), 
			0, NULL);

	clip->set_selection = clipboard_set_selection;
	clip->set_selection_owner = clipboard_set_selection_owner;

	clip->get_selection = clipboard_get_selection;
	clip->get_selection_owner = clipboard_get_selection_owner;

	clip->selection_clear = clipboard_selection_clear;
	clip->send_selection = clipboard_send_selection;
	clip->selection_notify = clipboard_selection_notify;
	clip->handle_incr_mechanism = clipboard_handle_incr_mechanism;

	return clip;
}

void clipboard_set_selection(px_clipboard* clip, const char* text)
{
	Display* display = display_get_Xdisplay();
	Window win;
	
	if (clip->text != NULL)
	{
		free(clip->text);
	}

	if (text != NULL)
		clip->text = strdup(text);
	else
		clip->text = NULL;
	win = XGetSelectionOwner(display, clip->selection);

	if (win != clip->win)
		XSetSelectionOwner(display, clip->selection, clip->win, CurrentTime);
}

void clipboard_set_selection_owner(px_clipboard* clip, px_widget* widget)
{
	if (clip->selection_owner != NULL && clip->selection_owner != widget)
	{
		px_event event;
		event.type = SELECTION_CLEAR;
		if (clip->selection_owner->default_callbacks[SELECTION_CLEAR] != NULL)
			clip->selection_owner->default_callbacks[SELECTION_CLEAR]
				(clip->selection_owner, &event);

		if (clip->selection_owner != NULL && 
			clip->selection_owner->callbacks[event.type] != NULL)
				clip->selection_owner->callbacks[event.type]
					(clip->selection_owner, &event);
	}

	clip->selection_owner = widget;
}

const char* clipboard_get_selection(px_clipboard* clip)
{
	Atom store;
	Window win;
	
	px_eventloop* eventloop = eventloop_get_eventloop();
	Display* display = display_get_Xdisplay();

	win = XGetSelectionOwner(display, clip->selection);
	
	if (win != clip->win)
	{
		if (clip->text != NULL)
			free(clip->text);

		clip->text = NULL;
	
		store = XInternAtom(display, "PICOENTRY_OUT", False);
		XConvertSelection(display, clip->selection, XA_STRING, store, clip->win, CurrentTime);
	
		eventloop->loop(eventloop);
		eventloop->run = 1;
	}

	return clip->text;
}

px_widget* clipboard_get_selection_owner(px_clipboard* clip)
{
	return clip->selection_owner;
}

void clipboard_selection_clear(px_clipboard* clip, px_event* event)
{
	if (clip->text != NULL)
	{
		free(clip->text);
		clip->text = NULL;
	}

	if (clip->selection_owner != NULL)
	{
		if (clip->selection_owner->default_callbacks[SELECTION_CLEAR] != NULL)
			clip->selection_owner->default_callbacks[SELECTION_CLEAR]
				(clip->selection_owner, event);

		if (clip->selection_owner != NULL && 
			clip->selection_owner->callbacks[event->type] != NULL)
				clip->selection_owner->callbacks[event->type]
					(clip->selection_owner, event);
	}
	clip->selection_owner = NULL;
}

void clipboard_send_selection(px_clipboard* clip, px_event* event)
{
	Window win = event->xev->xselectionrequest.requestor;
	Window store = event->xev->xselectionrequest.property;
	XEvent response;
	Display* display = display_get_Xdisplay();

	if (clip->text == NULL || clip->text[0] == '\0')
	{
		// Something wrong has happened ... there is no selection but I am the owner ...
		return;
	}

	printf("----------- Will send %s to %lu\n", clip->text, win);
	XChangeProperty(display, win, store, XA_STRING,	8, PropModeReplace, (unsigned char*) clip->text, (int)strlen(clip->text));
	
	// set values for the response event
	response.xselection.property = store;
	response.xselection.type = SelectionNotify;
	response.xselection.display = display;
	response.xselection.requestor = win;
	response.xselection.selection = event->xev->xselectionrequest.selection;
	response.xselection.target = event->xev->xselectionrequest.target;
	response.xselection.time = event->xev->xselectionrequest.time;

	// send the response event
	XSendEvent(display, win, 0, 0, &response);
	XFlush(display);
	return;
}

void clipboard_selection_notify(px_clipboard* clip, px_event* event)
{
	Display* display = display_get_Xdisplay();
	Atom inc = XInternAtom(display, "INCR", False);
	Atom store = XInternAtom(display, "PICOENTRY_OUT", False);
	Atom store_type;

	int store_format;
	unsigned long store_size, store_items;
	unsigned char* buffer, *selection;
	px_eventloop* eventloop = eventloop_get_eventloop();

	// No selection available ... so dont do anything
	if (event->xev->xselection.property == None)
	{
		eventloop->exit(eventloop);
		return;
	}

	// find the size and format of the data in property 

	XGetWindowProperty(display, clip->win, store, 0, 0, False, AnyPropertyType, &store_type,
			&store_format, &store_items, &store_size, &buffer);
	XFree(buffer);

	if (store_type == inc)
	{
		// start INCR mechanism by deleting property
		XDeleteProperty(display, clip->win, store);
		XFlush(display);
		return;
	}

	// if it's not incr, and not format == 8, then there's nothing in the selection (that we understands, anyway)
	
	if (store_format == 8)
	{
		XGetWindowProperty(display, clip->win, store,	0, (long)store_size, False, AnyPropertyType,
				&store_type, &store_format, &store_items, &store_size, &buffer);
		
		// finished with property, delete it 
		XDeleteProperty(display, clip->win, store);
		
		// copy the buffer to the pointer for returned data 
		selection = (unsigned char *)malloc(sizeof(unsigned char) * (store_items+1));
		memcpy(selection, buffer, store_items);
		selection[store_items] = '\0';

		// free the buffer
		XFree(buffer);
			
		// Insert the text into the widget
		
		if (clip->text != NULL)
			free(clip->text);

		clip->text = strdup(selection);
	}
	else
	{
		XFree(buffer);
		XDeleteProperty(display, clip->win, store);
	}
	eventloop->exit(eventloop);
	return;
}

void clipboard_handle_incr_mechanism(px_clipboard* clip, px_event* event)
{
	Display* display = display_get_Xdisplay();
	if (event->xev->type == PropertyNotify)
	{
		/****************************************************************
		 * Implementation of the INCR method of obtaining selection	*
		 * To use the INCR method, we basically delete the property	* 
		 * with the selection in it, wait for an event indicating that	*
		 * the property has been created, then read it, delete it, etc.	*
		 *								*
		 * **************************************************************/

		Atom store = XInternAtom(display, "PICOENTRY_OUT", False);
		Atom store_type;

		int store_format;
		unsigned long store_size, store_items;
		unsigned char* buffer, *selection;

		// skip unless the property has a new value

		if (event->xev->xproperty.state == PropertyNewValue)
		{
			// check size and format of the property
			XGetWindowProperty(display, clip->win, store, 0, 0, False, AnyPropertyType, 
					&store_type, &store_format, &store_items, &store_size, (unsigned char **)&buffer);

			if (store_format == 8)
			{
				if (store_size == 0)
				{
					// no more data, exit from loop
					px_eventloop* eventloop = eventloop_get_eventloop();
					eventloop->exit(eventloop);

					XFree(buffer);
					XDeleteProperty(display, clip->win, store);
				}
				else
				{
					// if we have come this far, the propery contains text, we know the size.
			 
					XGetWindowProperty(display, clip->win, store, 0, (long)store_size,
							False, AnyPropertyType, &store_type, &store_format,
							&store_items, &store_size, (unsigned char **)&buffer);

					// copy the buffer to the pointer for returned data 
					selection = (unsigned char *)malloc(sizeof(unsigned char) * (store_items+1));
					memcpy(selection, buffer, store_items);
					selection[store_items] = '\0';
					
					if (clip->text != NULL)
					{
						char* t = NULL;
						t = (char*)malloc(sizeof(char) * (strlen(clip->text) + strlen(selection) + 1));
						strcpy(t, clip->text);
						strcat(t, selection);
					}
					else
						clip->text = strdup(selection);
					
					
							
					free(selection);

					// delete property to get the next item 
					XDeleteProperty(display, clip->win, store);
					XFlush(display);
				}
			}
		}
		else
		{
			// property does not contain text, delete it to tell the other X client that we have read
			//  it and to send the next property
			px_eventloop* eventloop = eventloop_get_eventloop();
			eventloop->exit(eventloop);
			
			XFree(buffer);
			XDeleteProperty(display, clip->win, store);
		}
	}

	return;
}
